# -*- coding: utf-8 -*-
import inspect
import os
from functools import partial, total_ordering

from yurlungur.core.wrapper import (
    YException, _YObject, _YAttr, OM
)
from yurlungur.tool.math import (
    YVector, YColor, YMatrix
)
from yurlungur.tool.meta import meta
from yurlungur.tool.util import trace


class YObject(_YObject):
    """base class
    >>> obj = YObject("pCone")
    >>> obj("cone")
    """

    def __init__(self, item):
        self.item = item

    def __repr__(self):
        return self.name

    def __call__(self, *args, **kwargs):
        if hasattr(meta, "rename"):
            return meta.rename(self.item, *args, **kwargs)

        if hasattr(meta, "root"):
            return meta.node(self.item).setName(*args, **kwargs)

        if hasattr(meta, "data"):
            meta.data.objects[self.item].name = "".join(args)

    @property
    def name(self):
        return self.item

    def parent(self, *args, **kwarg):
        if len(args) > 0:
            return meta.parent(self.item, *args, **kwarg)

        elif len(args) == 0 and len(kwarg) > 0:
            return meta.parent(self.item, *args, **kwarg)

        else:
            if hasattr(meta, "getAttr"):
                return YObject(
                    partial(meta.listRelatives, self.item, p=1)(*args, **kwarg)
                )

            if hasattr(meta, "root"):
                return YObject(
                    meta.node(self.item).parent().name()
                )

    def instance(self, *args, **kwarg):
        if hasattr(meta, "instance"):
            if len(args) > 0:
                return meta.instance(self.item, lf=1)
            else:
                return meta.listRelatives(self.item, ap=1, f=1)[1:] or None

        raise YException

    def select(self, *args, **kwargs):
        if 'shape' not in kwargs and 's' not in kwargs:
            kwargs['s'] = True

        if hasattr(meta, "select"):
            if len(args) == 0 and len(kwargs) == 0:
                meta.select(*args, **kwargs)
            else:
                meta.ls(sl=1)

        raise YException

    @trace
    def delete(self, *args, **kwargs):
        if hasattr(meta, "delete"):
            return meta.delete(self.name, *args, **kwargs)

        if hasattr(meta, "root"):
            return meta.node(self.name).destroy()

        if hasattr(meta, "context"):
            return meta.context.scene.objects.unlink(meta.data.objects[self.name])

        raise YException

    @property
    def children(self, *args, **kwarg):
        if hasattr(meta, "getAttr"):
            return partial(meta.listRelatives, self.item, c=1)(*args, **kwarg) or None

        raise YException

    @trace
    def hide(self, on=True):
        if hasattr(meta, "data"):
            meta.data.objects[self.name].hide = on

        raise YException

    @trace
    def attr(self, val, *args, **kwargs):
        if hasattr(meta, "getAttr"):
            return YAttr(
                meta.getAttr(self.name + "." + val, *args, **kwargs), self.name, val
            )

        if hasattr(meta, "root"):
            parm = (meta.node(self.name).parm(val)
                    or meta.node(self.name).parmTuple(val))
            return YAttr(
                parm.eval(), self.name, val
            )

        if hasattr(meta, "data"):
            return YAttr(
                meta.data.objects[self.name].name, self.name, val
            )

        raise YException

    def __getattr__(self, item):
        if hasattr(meta, "getAttr"):
            return YAttr(
                meta.getAttr(self.name + "." + item), self.name, item
            )

        if hasattr(meta, "root"):
            parm = (meta.node(self.name).parm(item) or
                    meta.node(self.name).parmTuple(item))
            return YAttr(
                parm.eval(), self.name, item
            )

        if hasattr(meta, "data"):
            return YAttr(
                meta.data.objects[self.name].name, self.name, item
            )

        raise YException

    @property
    def attrs(self, *args, **kwargs):
        if hasattr(meta, "listAttr"):
            return tuple(meta.listAttr(self.name, *args, **kwargs)) or None

        if hasattr(meta, "root"):
            return tuple(
                p.name() for p in meta.node(self.name).parms() or []
            )

        if hasattr(meta, "data"):
            return inspect.getmembers(meta.data.objects[self.name])

        raise YException

    @trace
    def geometry(self):
        if hasattr(meta, "ls"):
            dag = OM.MGlobal.getSelectionListByName(self.name).getDagPath(0)
            return OM.MFnMesh(dag)

        if hasattr(meta, "root"):
            return meta.node(self.name).geometry()

        if hasattr(meta, "data"):
            return meta.data.meshes[self.name]

        raise YException

    @property
    def id(self):
        if hasattr(meta, "ls"):
            return meta.ls(self.name, uuid=1)[0] or 0

        if hasattr(meta, "root"):
            return meta.node(self.name).sessionId() or 0

        if hasattr(meta, "data"):
            return meta.data.objects[self.name].id_data or 0

        raise YException


class YNode(YObject):
    """connect-able object"""

    def __init__(self, item=None):
        super(YNode, self).__init__(item)
        self.item = item

    @trace
    def create(self, *args, **kwargs):
        if hasattr(meta, "createNode"):
            return YObject(meta.createNode(*args, **kwargs))

        if hasattr(meta, "root"):
            if len(args) == 0 and len(kwargs) == 0:
                return YNode(
                    partial(
                        meta.node(self.name).createNode, self.name
                    )(*args, **kwargs).path())
            return YNode(meta.node(self.name).createNode(*args, **kwargs).path())

        if hasattr(meta, "ops"):
            try:
                getattr(meta.ops.mesh, str(self).lower() + "_add")(*args, **kwargs)
            except AttributeError:
                getattr(meta.ops.object, str(self).lower() + "_add")(*args, **kwargs)

        raise YException

    @trace
    def connect(self, *args, **kwargs):
        if hasattr(meta, "root"):
            return partial(
                meta.node(self.name).setInput, 0)(*args, **kwargs)

        raise YException

    @trace
    def disconnect(self, *args, **kwargs):
        if hasattr(meta, "root"):
            return partial(
                meta.node(self.name).setInput, 0, None)(*args, **kwargs)

        raise YException

    def inputs(self, *args, **kwargs):
        if hasattr(meta, "listConnections"):
            return partial(meta.listConnections, s=1)(*args, **kwargs)

        if hasattr(meta, "root"):
            return meta.node(self.name).inputs()

        raise YException

    def outputs(self, *args, **kwargs):
        if hasattr(meta, "listConnections"):
            return partial(meta.listConnections, d=1)(*args, **kwargs)

        if hasattr(meta, "root"):
            return meta.node(self.name).outputs()

        raise YException


@total_ordering
class YAttr(_YAttr):
    """parametric object"""

    def __init__(self, *args, **kwargs):
        assert len(args) > 2, "parameter is invalid."
        self._values = args
        self.obj, self.val = self._values[1:]

    def __getitem__(self, idx):
        return self._values[idx]

    def __repr__(self):
        return str(self.value[0])

    def __eq__(self, other):
        return self.value[0] == other.value[0]

    def __gt__(self, other):
        return self.value[0] >= other.value[0]

    @property
    def value(self):
        return self._values[0]

    @trace
    def set(self, *args, **kwargs):
        if hasattr(meta, "setAttr"):
            return meta.setAttr(self.obj + "." + self.val, *args, **kwargs)

        if hasattr(meta, "root"):
            parm = (meta.node(self.obj).parm(self.val) or
                    meta.node(self.obj).parmTuple(self.val)
                    )
            return parm.set(
                args[0].tolist() if hasattr(args[0], "T") else args[0], **kwargs)

        if hasattr(meta, "data"):
            return setattr(meta.data.objects[self.obj],
                           self.val, args[0].tolist() if hasattr(args[0], "T") else args[0])

        raise YException

    @trace
    def lock(self, on):
        if hasattr(meta, "setAttr"):
            return meta.setAttr(self.obj + "." + self.val, lock=on)

        if hasattr(meta, "root"):
            return meta.node(self.obj).parm(self.val).lock(on)

        raise YException

    @trace
    def hide(self, on=True):
        if hasattr(meta, "setAttr"):
            return meta.setAttr(self.obj + "." + self.val, keyable=not on, channelBox=not on)

        if hasattr(meta, "root"):
            return meta.node(self.obj).parm(self.val).hide(on)

        raise YException

    @trace
    def connect(self, val, **kwargs):
        if hasattr(meta, "connectAttr"):
            return partial(
                meta.connectAttr, self.obj + "." + self.val
            )(val[1] + "." + val[2], **kwargs)

        raise YException

    @trace
    def disconnect(self, val, **kwargs):
        if hasattr(meta, "disconnectAttr"):
            return partial(
                meta.disconnectAttr, self.obj + "." + self.val
            )(val[1] + "." + val[2], **kwargs)

        raise YException

    @property
    def asVector(self):
        try:
            return YVector(self._values[0])
        except TypeError:
            return YVector(*self._values[0])

    @property
    def asColor(self):
        try:
            return YColor(self._values[0])
        except TypeError:
            return YColor(*self._values[0])

    @property
    def asMatrix(self):
        try:
            return YMatrix(self._values[0])
        except TypeError:
            return YMatrix(*self._values[0])


class YFile(object):
    """save, load and export"""

    def __init__(self, file=""):
        self.file = file

    @classmethod
    def load(cls, *args, **kwargs):
        if hasattr(meta, "file"):
            return cls(partial(meta.file, i=1)(*args, **kwargs))

        if hasattr(meta, "hipFile"):
            return cls(meta.hipFile.load(*args, **kwargs))

        if hasattr(meta, "ops"):
            return cls(meta.ops.wm.open_mainfile(*args, **kwargs))

        raise YException

    @classmethod
    def save(cls, *args, **kwargs):
        if hasattr(meta, "file"):
            return cls(partial(meta.file, s=1)(*args, **kwargs))

        if hasattr(meta, "hipFile"):
            return cls(meta.hipFile.save(*args, **kwargs))

        if hasattr(meta, "ops"):
            return meta.ops.wm.save_mainfile(*args, **kwargs)

        raise YException

    @property
    def current(cls):
        if hasattr(meta, "file"):
            return cls(meta.file(exn=1, q=1))

        if hasattr(meta, "hipFile"):
            return cls(meta.hipFile.path())

        if hasattr(meta, "data"):
            return cls(meta.data.filepath)

        raise YException

    def reference(cls):
        raise YException

    @property
    def name(self):
        return os.path.basename(self.file)

    @property
    def path(self):
        return os.path.dirname(self.file)



# -*- coding: utf-8 -*-
from yurlungur.core import env
from yurlungur.core.proxy import YNode, YFile
from yurlungur.tool.meta import meta

__all__ = ["file", "cmd", "UndoGroup"]


class UndoGroup(object):
    def __init__(self, label):
        self.label = label

    def __enter__(self):
        if env.Maya():
            meta.undoInfo(ock=1)
            return self
        elif env.Blender():
            self.undo = meta.context.user_preferences.edit.use_global_undo
            meta.context.user_preferences.edit.use_global_undo = False

    def __exit__(self, exc_type, exc_value, traceback):
        if env.Maya():
            meta.undoInfo(cck=1)
        elif env.Blender():
            meta.context.user_preferences.edit.use_global_undo = self.undo


class Command(object):
    @staticmethod
    def register(func):
        pass

    @staticmethod
    def unregister(func):
        pass

    @classmethod
    def list(cls):
        return [obj for obj in dir(cls) if not obj.startswith("_")]


def _ls(cls, *args, **kwargs):
    gen = meta.ls(*args, **kwargs) if hasattr(meta, "ls") else meta.pwd().allItems()
    return tuple(YNode(obj) for obj in gen)


def _rm(cls, *args):
    for obj in args:
        YNode(obj).delete()


def _glob(cls, *args, **kwargs):
    gen = meta.ls(*args, **kwargs) if hasattr(meta, "ls") else meta.pwd().glob(*args, **kwargs)
    return tuple(YNode(obj) for obj in gen)


def _select(cls, *args, **kwargs):
    if hasattr(meta, "select"):
        meta.select(*args, **kwargs)

    if hasattr(meta, "root"):
        for node in meta.nodes(args):
            node.setSelected(True, **kwargs)


def _alembicImporter(cls, *args, **kwargs):
    """
    >>> f = YFile()
    >>> YFile.new_method = new_method
    >>> print f.new_method("new")
    """
    if hasattr(meta, "AbcImport"):
        return cls(meta.AbcImport(*args, **kwargs))

    if hasattr(meta, "root"):
        geo = yr.YNode("obj").create("geo")
        abc = geo.create("alembic")
        abc.fileName.set(*args)
        return cls(*args)


def _alembicExporter(cls, *args, **kwargs):
    if hasattr(meta, "AbcExport"):
        return cls(meta.AbcExport(*args, **kwargs))


def _fbxImporter(cls, *args, **kwargs):
    if hasattr(meta, "importFBX"):
        return meta.importFBX(*args, **kwargs)
    else:
        return cls(meta.eval("FBXImport -file {0};".format(*args)))


def _fbxExporter(cls, *args, **kwargs):
    return cls(meta.eval("FBXExportInAscii -v true; FBXExport -f \"{}\" -s;".format(*args)))


file = YFile()
cmd = Command()

# Monkey-Patch
if env.Maya():
    for plugin in "fbxmaya.mll", "AbcImport.mll", "AbcExport.mll":
        meta.loadPlugin(plugin, qt=1)

    YFile.abcImporter = _alembicImporter
    YFile.abcExporter = _alembicExporter
    YFile.fbxImporter = _fbxImporter
    YFile.fbxExporter = _fbxExporter

    Command.ls = _ls
    Command.rm = _rm
    Command.glob = _glob
    Command.select = _select

if env.Houdini():
    UndoGroup = meta.undos.group

    YFile.abcImporter = _alembicImporter
    YFile.fbxImporter = _fbxImporter

    Command.ls = _ls
    Command.rm = _rm
    Command.glob = _glob
    Command.select = _select

if env.Blender():
    YFile.abcImporter = _alembicImporter
    YFile.fbxImporter = _fbxImporter

if env.Unreal():
    UndoGroup = meta.ScopedEditorTransaction



# waifu2x.py by @marcan42 based on https://github.com/nagadomi/waifu2x
# MIT license, see https://github.com/nagadomi/waifu2x/blob/master/LICENSE
import sys
import json

import numpy as np
from scipy import misc, signal
from PIL import Image

# Model export script: https://mrcn.st/t/export_model.lua (needs a working waifu2x install)
def main():
    infile, outfile, modelpath = sys.argv[1:]
    model = json.load(open(modelpath))
    im = Image.open(infile).convert("YCbCr")
    im = misc.fromimage(im.resize((2*im.size[0], 2*im.size[1]), resample=Image.NEAREST)).astype("float32")
    planes = [np.pad(im[:,:,0], len(model), "edge") / 255.0]
    count = sum(step["nInputPlane"] * step["nOutputPlane"] for step in model)
    progress = 0

    for step in model:
        assert step["nInputPlane"] == len(planes)
        assert step["nOutputPlane"] == len(step["weight"]) == len(step["bias"])
        o_planes = []
        for bias, weights in zip(step["bias"], step["weight"]):
            partial = None
            for ip, kernel in zip(planes, weights):
                p = signal.convolve2d(ip, np.float32(kernel), "valid")
                if partial is None:
                    partial = p
                else:
                    partial += p
                progress += 1
                sys.stderr.write("\r%.1f%%..." % (100 * progress / float(count)))

            partial += np.float32(bias)
            o_planes.append(partial)

        planes = [np.maximum(p, 0) + 0.1 * np.minimum(p, 0) for p in o_planes]

    assert len(planes) == 1
    im[:,:,0] = np.clip(planes[0], 0, 1) * 255
    misc.toimage(im, mode="YCbCr").convert("RGB").save(outfile)
    sys.stderr.write("Done\n")

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
from __future__ import print_function

try:
    import __builtin__
except ImportError:
    import builtins as __builtins__

import sys
import time
from logging import (
    getLogger, Handler, INFO, WARNING, basicConfig
)
from pprint import pformat

import yurlungur
from yurlungur.core import env


class GuiLogHandler(Handler):
    def __init__(self, *args, **kwargs):
        super(GuiLogHandler, self).__init__(*args, **kwargs)

        if env.Maya():
            import maya.OpenMaya as om
            self.MGlobal = om.MGlobal

    def emit(self, record):
        from yurlungur.tool.meta import meta

        msg = self.format(record)
        if record.levelno > WARNING:
            if env.Maya():
                self.MGlobal.displayError(msg)
            elif env.Houdini():
                meta.ui.setStatusMessage(msg, severity=meta.severityType.Error)
            elif env.Unreal():
                meta.log_error(msg)
            elif env.Unity():
                meta.Debug.LogError(msg)

        elif record.levelno > INFO:
            if env.Maya():
                self.MGlobal.displayWarning(msg)
            elif env.Houdini():
                meta.ui.setStatusMessage(msg, severity=meta.severityType.Warning)
            elif env.Unreal():
                meta.log_warning(msg)
            elif env.Unity():
                meta.Debug.LogWarning(msg)

        else:
            if env.Maya():
                self.MGlobal.displayInfo(msg)
            elif env.Houdini():
                meta.ui.setStatusMessage(msg, severity=meta.severityType.Message)
            elif env.Unreal():
                meta.log(msg)
            elif env.Unity():
                meta.Debug.Log(msg)


# logger
logger = getLogger(yurlungur.__name__)
logger.setLevel(INFO)

# TODO: https://code.blender.org/2016/05/logging-from-python-code-in-blender/
if not env.Blender():
    handler = GuiLogHandler()
    logger.addHandler(handler)
basicConfig(level=INFO, stream=sys.stdout)


def log(*msgs):
    logger.info(pformat(*msgs))


def print(*args):
    __builtin__.print(pformat(*args))


def _progress():
    for i in range(1, 101):
        sys.stdout.flush()
        log = "\r%d%% (%d of %d)" % (i, i, 100)
        sys.stdout.write(log + "")
        sys.stdout.flush()
        time.sleep(0.01)



# -*- coding: utf-8 -*-
from yurlungur.core.app import application
from yurlungur.core.proxy import YNode, YAttr
from yurlungur.tool.math import YVector, YMatrix, YColor
from yurlungur.core.env import Maya, Houdini, Unreal
from yurlungur.core.wrapper import OM


class Builder(object):
    """
    http://help.autodesk.com/view/MAYAUL/2017/JPN/?guid=__py_ref_class_open_maya_1_1_m_fn_mesh_html
    http://www.sidefx.com/docs/houdini/hom/hou/Geometry.html
    """

    def geometry(self):
        pass

    def camera(self):
        pass

    def light(self):
        pass


class Shader(object):
    def material(self):
        pass

    def texture(self):
        pass

    def renderer(self):
        pass



# -*- coding: utf-8 -*-
import inspect

import yurlungur
from yurlungur.core import app, env

"""internal module"""


class YException(NotImplementedError):
    """
    >>> raise NotImplementedError(app.application)
    """
    pass


class YMObject(object):
    """command wrapper for any application"""

    def __getattr__(self, item):
        for cmd, _ in inspect.getmembers(app.application):
            if cmd == item:
                setattr(
                    yurlungur, cmd,
                    (lambda str: dict(inspect.getmembers(app.application))[str])(cmd)
                )
                return getattr(yurlungur, item)

        return None

    def eval(self, script):
        if env.Maya():
            import maya.mel as mel
            return mel.eval(script)
        if env.Houdini():
            app.application.hscript(script)

        raise YException

    @property
    def module(self):
        return app.application.__name__


class MetaObject(type):
    def __new__(cls, name, bases, attrs):
        return super(MetaObject, cls).__new__(cls, name, bases, attrs)


class MetaAttr(type):
    def __new__(cls, name, bases, attrs):
        return super(MetaAttr, cls).__new__(cls, name, bases, attrs)


class MetaNode(type):
    def __new__(cls, name, bases, attrs):
        return super(MetaNode, cls).__new__(cls, name, bases, attrs)


# Dynamic Class
_YObject = MetaObject("YObject", (object,), {"__doc__": MetaObject.__doc__})
_YNode = MetaNode("YNode", (object,), {"__doc__": MetaNode.__doc__})
_YAttr = MetaAttr("YAttr", (object,), {"__doc__": MetaAttr.__doc__})
_YVector = _YMatrix = _YColor = OM = object

if env.Maya():
    import maya.api.OpenMaya as OM

    _YVector = type('_YVector', (OM.MVector,), dict())
    _YMatrix = type('_YMatrix', (OM.MMatrix,), dict())
    _YColor = type('_YColor', (OM.MColor,), dict())

elif env.Houdini() or env.Unreal():
    meta = YMObject()

    _YVector = type('_YVector', (
        meta.Vector if hasattr(meta, "Vector") else meta.Vector3,
    ), dict())

    _YMatrix = type('_YMatrix', (
        meta.Matrix if hasattr(meta, "Matrix") else meta.Matrix4,
    ), dict())

    _YColor = type('_YColor', (meta.Color,), dict())

elif env.Blender():
    import mathutils

    _YVector = type('_YVector', (mathutils.Vector,), dict())
    _YMatrix = type('_YMatrix', (mathutils.Matrix,), dict())
    _YColor = type('_YColor', (mathutils.Color,), dict())
